//
// Tools for Harris Corner Detector
//

#include	<cmath>

#include	"HarrisTools.h"

// make Gauss window
void
make_Gauss_window(double Gw[], int Ns, double sigma)
{
   int		x, v0, u0;
   double	d;
   
   x = 0;
   for (v0 = -Ns; v0 <= Ns; v0++)
      for (u0 = -Ns; u0 <= Ns; u0++)
      {
	 d = exp( - (sqr((double)u0) + sqr((double)v0)) / (2.0 * sqr(sigma)));
	 Gw[x++] = d;
      }
}

// check value greater than or equal upper left area
int
check_ge_upperleft(double d, float out[], int x, int width)
{
   int	w2 = width * 2;

   if (d >= out[x - w2 - 2] &&
       d >= out[x - w2 - 1] &&
       d >= out[x - w2]     &&
       d >= out[x - w2 + 1] &&
       d >= out[x - w2 + 2] &&

       d >= out[x - width - 2] &&
       d >= out[x - width - 1] &&
       d >= out[x - width]     &&
       d >= out[x - width + 1] &&
       d >= out[x - width + 2] &&

       d >= out[x - 2] &&
       d >= out[x - 1])
      return 1;

   return 0;
}

// check value lessr than upper left area
int
check_gt_lowerright(double d, float out[], int x, int width)
{
   int	w2 = width * 2;

   if (d >  out[x + 1] &&
       d >  out[x + 2] &&

       d >  out[x + width - 2] &&
       d >  out[x + width - 1] &&
       d >  out[x + width]     &&
       d >  out[x + width + 1] &&
       d >  out[x + width + 2] &&

       d >  out[x + w2 - 2] &&
       d >  out[x + w2 - 1] &&
       d >  out[x + w2] &&
       d >  out[x + w2 + 1] &&
       d >  out[x + w2 + 2])
      return 1;

   return 0;
}

// corner list
int
add_list(int u, int v, double d, int *up, int *vp, double *dp,
	 int Num, int MaxNum, double MinDist)
{
   int	i, j, k;
   double	dist;

   for (i = 0; i < Num; i++)
   {
      if (((dist = distance(u, v, up[i], vp[i])) < MinDist)
	  && (d <= dp[i]))
      {
	 return Num;
      }

      if (d > dp[i]) break;
   }

   if (i == Num) // add the point to the last of the list
   {
      if (Num < MaxNum)
      {
	 Num++;
	 up[i] = u;
	 vp[i] = v;
	 dp[i] = d;
      }
   }
   else // insert the point to the list
   {
      if (Num < MaxNum)
	 Num++;

      for (j = Num - 1; j > i; j--) // shift the list follow inserting point
      {
	 up[j] = up[j-1];
	 vp[j] = vp[j-1];
	 dp[j] = dp[j-1];
      }
      up[i] = u;
      vp[i] = v;
      dp[i] = d;
      for (j = k = i + 1; k < Num; k++)
      {
	 while (k < Num && (dist = distance(u, v, up[k], vp[k])) < MinDist)
	    k++;

	 if (k < Num)
	 {
	    up[j] = up[k];
	    vp[j] = vp[k];
	    dp[j] = dp[k];
	    j++;
	 }
      }
      Num = j;
   }

   return Num;
}
