//
//	ưбդ version 2.2.2
//	ᥤ
//

#define	MAIN

#include	<iostream>
#include	<fstream>

using namespace std;

#include	<unistd.h>

#include	"CImage.h"
#include	"Image.h"
#include	"Mat.h"
#include	"Corresp_cls.h"
#include	"init_match.h"
#include	"automatch.h"
#include	"matcher_consts.h"

double	EPS = 1e-10;
double	DET_EPS = 1e-10;

// form ؿ
#define TMPBUFLEN       80
static char tmpbuff[TMPBUFLEN];

void
usage(char *command)
{
   cerr << "usage: " << command
	<< " [-f f0] [-P Pmax] [-t tpsize] [-1 image1-point -2 image2-point]"
	<< " [-M matrix-file] [-m match-file] "
	<< " [-d distance] [-H] [-n] [-v] [-h] image1-ppm image2-ppm" << endl;
   cerr << "  Input file:  image1-ppm, image2-ppm" << endl;
   cerr << "  Output file: image1-point, image2-point, matrix-file, match-file" << endl;
   cerr << "  Miscellaneous options:" << endl;
   cerr << "     -H	using starndard Harris instead of improved Harris" << endl;
   cerr << "     -n	doing normalized correlation" << endl;
   cerr << "     -h	help" << endl;
   cerr << "     -v	verbose" << endl;
}

void
output_matches(char *filename, CrpPair *cp, int N, Point *pa, Point *pb)
{
   if (*filename == '\0')
   {
      cout << N << endl;
      for (int i = 1; i <= N; i++)
      {
         snprintf(tmpbuff, TMPBUFLEN, "(%4d,%4d)--(%4d,%4d)",
                  pa[cp[i].a].u, pa[cp[i].a].v,
                  pb[cp[i].b].u, pb[cp[i].b].v);
	 cout << tmpbuff << endl;
      }
   }
   else
   {
      ofstream	fout(filename);

      if (fout.bad())
      {
	 cerr << "output_matches(): unable to open " << filename << endl;
	 return;
      }
      for (int i = 1; i <= N; i++)
      {
         snprintf(tmpbuff, TMPBUFLEN, "(%4d,%4d)--(%4d,%4d)",
                  pa[cp[i].a].u, pa[cp[i].a].v,
                  pb[cp[i].b].u, pb[cp[i].b].v);
	 fout << tmpbuff << endl;
      }
      fout.close();
   }
}

void
output_features(char *filename, Point *pt, int N)
{
   ofstream	fout(filename);

   if (fout.bad())
   {
      cerr << "output_features(): unable to open " << filename << endl;
      return;
   }

   for (int i = 0; i < N; i++)
   {
      snprintf(tmpbuff, TMPBUFLEN, "(%4d,%4d)", pt[i].u, pt[i].v);
      fout << tmpbuff << endl;
   }

   fout.close();
}
   
void
output_mat(char *filename, Mat& A)
{
   ofstream	fout(filename);

   if (fout.bad())
   {
      cerr << "output_mat(): unable to open " << filename << endl;
      return;
   }

   snprintf(tmpbuff, TMPBUFLEN, "(%20.15f %20.15f %20.15f )",
            A[0][0], A[0][1], A[0][2]);
   fout << tmpbuff << endl;
   snprintf(tmpbuff, TMPBUFLEN, "(%20.15f %20.15f %20.15f )",
            A[1][0], A[1][1], A[1][2]);
   fout << tmpbuff << endl;
   snprintf(tmpbuff, TMPBUFLEN, "(%20.15f %20.15f %20.15f )",
            A[2][0], A[2][1], A[2][2]);
   fout << tmpbuff << endl;
}
   

int
main(int argc, char **argv)
{
   CImage	CIm1, CIm2;
   CrpPair	*cp;
   Point	*pa, *pb;
   int		N, Na, Nb;
   char		im1_name[80], im2_name[80], match_name[80];
   char		M_name[80], im1pt_name[80], im2pt_name[80];
   int		max_corner, tp_size, Fflag;
   int		verbose, normflag, ImpHrsFlg;
   double	d, fl;
   Mat		F, H, A;
   int		opt;

   // ǥեͤΥå
   fl = DEFAULT_FL;
   max_corner = DEFAULT_CNRNUM;
   tp_size = DEFAULT_TPSIZE;
   d = DEFAULT_DIST;
   verbose = VERBOSE;
   normflag = 0;
   ImpHrsFlg = 1;
   im1pt_name[0] = im2pt_name[0] = M_name[0] = match_name[0] = '\0';

   // ץ
   while ((opt = getopt(argc, argv, "1:2:M:f:P:t:d:m:Hvhn")) != -1)
   {
      switch (opt)
      {
	case '1':
	   strcpy(im1pt_name, optarg);
	   break;
	case '2':
	   strcpy(im2pt_name, optarg);
	   break;
	case 'M':
	   strcpy(M_name, optarg);
	   break;
	case 'f':
	   fl = atof(optarg);
	   break;
	case 'P':
	   max_corner = atoi(optarg);
	   break;
	case 't':
	   tp_size = atoi(optarg);
	   if (!(tp_size % 2))
	      cerr << "Warning: we use " << (tp_size * 2 + 1)
		   << " as template size!" << endl;
	   tp_size /= 2;
	   break;
	case 'd':
	   d = atof(optarg);
	   break;
	case 'm':
	   strcpy(match_name, optarg);
	   break;
	case 'h':
	   usage(argv[0]);
	   exit(1);
	   break;
	case 'v':
	   verbose = 1;
	   break;
	case 'n':
	   normflag = 1;
	   break;
	case 'H':
	   ImpHrsFlg = 0;
	   break;
      }
   }

   if (optind < argc)
   {
      strcpy(im1_name, argv[optind++]);
      strcpy(im2_name, argv[optind++]);
   }
   else
   {
      usage(argv[0]);
      exit(1);
   }

   if (verbose)
   {
      cerr << "[Inputs]" << endl;
      cerr << "  First Image: " << im1_name << endl;
      cerr << "  Second Image: " << im2_name << endl;
      cerr << "  Focal length: " << fl << endl;
      cerr << "  Maximum number of feature points: " << max_corner << endl;
      cerr << "  Template size: " << tp_size * 2 + 1
	   << " x " << tp_size * 2 + 1  << endl;
      cerr << "  Threshold of distance: " << d << endl;
      cerr << "  Template matching method: " << ((normflag)? "normalized": "raw data") << endl;
      cerr << "[Outputs]" << endl;
      cerr << "  Matches => " << match_name << endl;
      cerr << "  Fundamental or homography matrix => " << M_name << endl;
      cerr << "  Feature points in the first image => " << im1pt_name << endl;
      cerr << "  Feature points in the second image => " << im2pt_name << endl;
      cerr << endl;
   }

   // եɤ߹
   load_ppm_file(CIm1, im1_name);
   load_ppm_file(CIm2, im2_name);

   // ưбդ
   Fflag = automatcher(CIm1, CIm2, d, fl, max_corner, tp_size,
		       A, &cp, &N, &pa, &Na, &pb, &Nb,
                       ImpHrsFlg, normflag, verbose);

   if (Fflag < 0)
      exit(1);

   output_matches(match_name, cp, N, pa, pb);

   if (im1pt_name[0] != '\0')
      output_features(im1pt_name, pa, Na);

   if (im2pt_name[0] != '\0')
      output_features(im2pt_name, pb, Nb);

   if (M_name[0] != '\0')
      output_mat(M_name, A);

   if (Fflag)
   {
      F = A;
      cout << "Fundamental matrix was selected!" << endl;
      cout << "F = ";
   }
   else
   {
      H = A;
      cout << "Homography matrix was selected!" << endl;
      cout << "H = ";
   }

   cout << endl;
   snprintf(tmpbuff, TMPBUFLEN, "(%20.15f %20.15f %20.15f )",
            A[0][0], A[0][1], A[0][2]);
   cout << tmpbuff << endl;
   snprintf(tmpbuff, TMPBUFLEN, "(%20.15f %20.15f %20.15f )",
            A[1][0], A[1][1], A[1][2]);
   cout << tmpbuff << endl;
   snprintf(tmpbuff, TMPBUFLEN, "(%20.15f %20.15f %20.15f )",
            A[2][0], A[2][1], A[2][2]);
   cout << tmpbuff << endl;
}
